#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Frame Check Sequence

The FCS is a sequence of 16 bits used for checking the integrity of a
received frame.

Derived from https://github.com/casebeer/afsk

Copyright (c) 2013 Christopher H. Casebeer. All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

   1. Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation
      and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

import struct

import bitarray

__author__ = 'Christopher H. Casebeer'  # NOQA pylint: disable=R0801
__copyright__ = 'Copyright (c) 2013 Christopher H. Casebeer. All rights reserved.'  # NOQA pylint: disable=R0801
__license__ = 'BSD 2-clause Simplified License'  # NOQA pylint: disable=R0801


class FCS(object):

    def __init__(self) -> None:
        self.fcs = 0xFFFF

    def update_bit(self, bit) -> None:
        check = (self.fcs & 0x1 == 1)
        self.fcs >>= 1
        if check != bit:
            self.fcs ^= 0x8408

    def update(self, bytes) -> None:
        for byte in (ord(b) for b in bytes):
            for i in range(7,-1,-1):
                self.update_bit((byte >> i) & 0x01 == 1)

    def digest(self):
#        print ~self.fcs
#        print "%r" % struct.pack("<H", ~self.fcs % 2**16)
#        print "%r" % "".join([chr((~self.fcs & 0xff) % 256), chr((~self.fcs >> 8) % 256)])
        # digest is two bytes, little endian
        return struct.pack("<H", ~self.fcs % 2**16)


def fcs(bits):
    """
    Append running bitwise FCS CRC checksum to end of generator
    """
    fcs = FCS()
    for bit in bits:
        yield bit
        fcs.update_bit(bit)

#    test = bitarray()
#    for byte in (digest & 0xff, digest >> 8):
#        print byte
#        for i in range(8):
#            b = (byte >> i) & 1 == 1
#            test.append(b)
#            yield b

    # append fcs digest to bit stream

    # n.b. wire format is little-bit-endianness in addition to little-endian
    digest = bitarray(endian="little")
    digest.frombytes(fcs.digest())
    for bit in digest:
        yield bit


def fcs_validate(bits):
    buffer = bitarray()
    fcs = FCS()

    for bit in bits:
        buffer.append(bit)
        if len(buffer) > 16:
            bit = buffer.pop(0)
            fcs.update(bit)
            yield bit

    if buffer.tobytes() != fcs.digest():
        raise Exception("FCS checksum invalid.")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Python APRS Module Class Definitions."""

import itertools
import logging
import socket
import time
import typing

import pkg_resources
import requests

import aprs  # pylint: disable=R0801

__author__ = 'Greg Albrecht W2GMD <oss@undef.net>'  # NOQA pylint: disable=R0801
__copyright__ = 'Copyright 2017 Greg Albrecht and Contributors'  # NOQA pylint: disable=R0801
__license__ = 'Apache License, Version 2.0'  # NOQA pylint: disable=R0801


class Frame(object):

    """
    Frame Class.

    Defines the components of an AX.25/APRS Frame.
    """

    __slots__ = ['source', 'destination', 'path', 'info']

    _logger = logging.getLogger(__name__)  # pylint: disable=R0801
    if not _logger.handlers:  # pylint: disable=R0801
        _logger.setLevel(aprs.LOG_LEVEL)  # pylint: disable=R0801
        _console_handler = logging.StreamHandler()  # pylint: disable=R0801
        _console_handler.setLevel(aprs.LOG_LEVEL)  # pylint: disable=R0801
        _console_handler.setFormatter(aprs.LOG_FORMAT)  # pylint: disable=R0801
        _logger.addHandler(_console_handler)  # pylint: disable=R0801
        _logger.propagate = False  # pylint: disable=R0801

    def __init__(self, source: bytes=b'', destination: bytes=b'',
                 path: list=[], info: bytes=b'') -> None:
        self.source = aprs.parse_callsign(source)
        self.destination = aprs.parse_callsign(destination)
        # TODO: Add parse_path function
        self.path = path
        self.info = aprs.parse_info_field(info)

    def __repr__(self) -> str:
        """
        Returns a string representation of this Object.
        """
        full_path = [str(self.destination)]
        full_path.extend([str(p) for p in self.path])
        frame = "%s>%s:%s" % (
            self.source,
            ','.join(full_path),
            self.info
        )
        return frame

    def __bytes__(self) -> bytes:
        full_path = [bytes(self.destination)]
        full_path.extend([bytes(p) for p in self.path])
        frame = b"%s>%s:%s" % (
            bytes(self.source),
            b','.join(full_path),
            bytes(self.info)
        )
        return frame

    def set_source(self, source: typing.Union[str, bytes]) -> None:
        self.source = aprs.parse_callsign(source)

    def set_destination(self, destination: typing.Union[str, bytes]) -> None:
        self.destination = aprs.parse_callsign(destination)

    def set_path(self, path=[]) -> None:
        self.path = [aprs.parse_callsign(pth) for pth in path]

    def update_path(self, update: bytes) -> None:
        self.path.append(aprs.parse_callsign(update))

    def set_info(self, info: typing.Union[str, bytes]) -> None:
        self.info = aprs.parse_info_field(info)

    def encode_ax25(self) -> bytes:
        """
        Encodes an APRS Frame as AX.25.
        """
        encoded_frame = []
        encoded_frame.append(aprs.AX25_FLAG)
        encoded_frame.append(self.destination.encode_ax25())
        encoded_frame.append(self.source.encode_ax25())
        for path_call in self.path:
            encoded_frame.append(path_call.encode_ax25())
        encoded_frame.append(aprs.ADDR_INFO_DELIM)
        encoded_frame.append(bytes(self.info))

        fcs = aprs.FCS()
        for bit in encoded_frame:
            fcs.update_bit(bit)

        encoded_frame.append(fcs.digest())
        encoded_frame.append(aprs.AX25_FLAG)

        return b''.join(encoded_frame)


class Callsign(object):

    """
    Callsign Class.

    Defines parts of an APRS AX.25 Callsign.
    """

    _logger = logging.getLogger(__name__)  # pylint: disable=R0801
    if not _logger.handlers:  # pylint: disable=R0801
        _logger.setLevel(aprs.LOG_LEVEL)  # pylint: disable=R0801
        _console_handler = logging.StreamHandler()  # pylint: disable=R0801
        _console_handler.setLevel(aprs.LOG_LEVEL)  # pylint: disable=R0801
        _console_handler.setFormatter(aprs.LOG_FORMAT)  # pylint: disable=R0801
        _logger.addHandler(_console_handler)  # pylint: disable=R0801
        _logger.propagate = False  # pylint: disable=R0801

    __slots__ = ['callsign', 'ssid', 'digi']

    def __init__(self, callsign: bytes=b'', ssid: bytes=b'0',
                 digi: bool=False) -> None:
        self.callsign: bytes = callsign
        self.ssid: bytes = ssid
        self.digi: bool = digi

    def __repr__(self) -> str:
        _callsign = self.callsign.decode()
        _ssid = self.ssid.decode()
        call_repr = _callsign

        # Don't print callsigns with ssid 0.
        if _ssid:
            try:
                if int(_ssid) > 0:
                    call_repr = '-'.join([_callsign, _ssid])
            except ValueError:
                if _ssid != 0:
                    call_repr = '-'.join([_callsign, _ssid])

        # If callsign was digipeated, append '*'.
        if self.digi:
            return ''.join([call_repr, '*'])
        else:
            return call_repr

    def __bytes__(self) -> bytes:
        _callsign = self.callsign
        _ssid = self.ssid
        call_repr = _callsign

        # Don't print callsigns with ssid 0.
        if _ssid:
            try:
                if int(_ssid) > 0:
                    call_repr = b'-'.join([_callsign, _ssid])
            except ValueError:
                if _ssid != 0:
                    call_repr = b'-'.join([_callsign, _ssid])

        # If callsign was digipeated, append '*'.
        if self.digi:
            return b''.join([call_repr, b'*'])
        else:
            return call_repr

    def set_callsign(self, callsign: bytes) -> None:
        self.callsign = callsign

    def set_ssid(self, ssid: bytes=b'0') -> None:
        if isinstance(ssid, bytes):
            self.ssid = ssid
        else:
            self.ssid = bytes(str(ssid), 'UTF-8')

    def set_digi(self, digi: bool) -> None:
        self.digi = digi

    def encode_ax25(self) -> bytearray:
        """
        Encodes Callsign as AX.25.
        """
        _callsign = self.callsign
        encoded_callsign = []

        encoded_ssid = (int(self.ssid) << 1) | 0x60

        if self.digi:
            # _callsign = ''.join([_callsign, '*'])
            encoded_ssid |= 0x80

        # Pad the callsign to at least 6 characters.
        while len(_callsign) < 6:
            _callsign += b' '

        for pos in _callsign:
            encoded_callsign.append(bytes([pos << 1]))

        encoded_callsign.append(bytes([encoded_ssid]))

        return b''.join(encoded_callsign)


class APRS(object):

    """APRS Object."""

    _logger = logging.getLogger(__name__)  # pylint: disable=R0801
    if not _logger.handlers:  # pylint: disable=R0801
        _logger.setLevel(aprs.LOG_LEVEL)  # pylint: disable=R0801
        _console_handler = logging.StreamHandler()  # pylint: disable=R0801
        _console_handler.setLevel(aprs.LOG_LEVEL)  # pylint: disable=R0801
        _console_handler.setFormatter(aprs.LOG_FORMAT)  # pylint: disable=R0801
        _logger.addHandler(_console_handler)  # pylint: disable=R0801
        _logger.propagate = False  # pylint: disable=R0801

    def __init__(self, user: bytes, password: bytes=b'-1') -> None:
        if isinstance(user, str):
            user = bytes(user, 'UTF-8')
        if isinstance(password, str):
            password = bytes(password, 'UTF-8')
        self.user = user

        try:
            version = bytes(pkg_resources.get_distribution(  # NOQA pylint: disable=E1101
                'aprs').version, 'UTF-8')
        except:  # pylint: disable=W0702
            version = b'GIT'
        version_str = b'Python APRS Module v' + version

        self._auth = b' '.join(
            [b'user', user, b'pass', password, b'vers', version_str])

        self._full_auth = None
        self.interface = None
        self.use_i_construct = False

    def start(self):
        """
        Abstract method for starting connection to APRS-IS.
        """
        pass

    def send(self, message):
        """
        Abstract method for sending messages to APRS-IS.
        """
        pass

    def receive(self, callback=None, frame_handler=aprs.parse_frame):
        """
        Abstract method for receiving messages from APRS-IS.
        """
        pass


class TCP(APRS):

    """APRS-IS TCP Class."""

    def __init__(self, user: bytes, password: bytes, servers: bytes=b'',
                 aprs_filter: bytes=b'') -> None:
        super(TCP, self).__init__(user, password)
        servers = servers or aprs.APRSIS_SERVERS  # Unicode
        aprs_filter = aprs_filter or b'/'.join([b'p', user])  # Unicode
        if isinstance(aprs_filter, str):
            aprs_filter = bytes(aprs_filter, 'UTF-8')

        # Unicode
        self._full_auth = b' '.join([self._auth, b'filter', aprs_filter])

        self.servers = itertools.cycle(servers)
        self.use_i_construct = True
        self._connected = False

    def start(self):
        """
        Connects & logs in to APRS-IS.
        """
        while not self._connected:
            servers = next(self.servers)
            if b':' in servers:
                server, port = servers.split(b':')
                port = int(port)
            else:
                server = servers
                port = aprs.APRSIS_FILTER_PORT

            try:
                addr_info = socket.getaddrinfo(server, port)

                self.interface = socket.socket(*addr_info[0][0:3])

                # Connect
                self._logger.info(
                    "Connect To %s:%i", addr_info[0][4][0], port)

                self.interface.connect(addr_info[0][4])

                server_hello = self.interface.recv(1024)

                self._logger.info(
                    'Connect Result "%s"', server_hello.rstrip())

                # Auth
                self._logger.info(
                    "Auth To %s:%i", addr_info[0][4][0], port)

                _full_auth = self._full_auth + b'\n\r'

                self.interface.sendall(_full_auth)

                server_return = self.interface.recv(1024)
                self._logger.info(
                    'Auth Result "%s"', server_return.rstrip())

                self._connected = True
            except socket.error as ex:
                self._logger.exception(ex)
                self._logger.warn(
                    "Error when connecting to %s:%d: '%s'",
                    server, port, str(ex))
                time.sleep(1)

    def send(self, frame):
        """
        Sends frame to APRS-IS.

        :param frame: Frame to send to APRS-IS.
        :type frame: str
        """
        self._logger.info('Sending frame="%s"', frame)

        # Unicode Sandwich: Send bytes.
        _frame = bytes(frame + b'\n\r')

        return self.interface.send(_frame)

    def receive(self, callback=None, frame_handler=aprs.parse_frame):
        """
        Receives from APRS-IS.

        :param callback: Optional callback to deliver frame to.
        :type callback: func

        :returns: Nothing, but calls a callback with an Frame object.
        :rtype: None
        """
        self._logger.info(
            'Receive started with callback="%s" and frame_handler="%s"',
            callback, frame_handler)

        # Unicode Sandwich: Receive Bytes.
        recvd_data = bytes()

        try:
            while 1:
                recv_data = self.interface.recv(aprs.RECV_BUFFER)

                if not recv_data:
                    break

                recvd_data += recv_data

                self._logger.debug('recv_data="%s"', recv_data.strip())

                if recvd_data.endswith(b'\r\n'):
                    lines = recvd_data.strip().split(b'\r\n')
                    recvd_data = bytes()
                else:
                    lines = recvd_data.split(b'\r\n')
                    recvd_data = lines.pop(-1)

                for line in lines:
                    if line.startswith(b'#'):
                        if b'logresp' in line:
                            self._logger.debug('logresp="%s"', line)
                        # We log all received data anyway, so no need to log
                        # it here again:
                        # else:
                        #    self._logger.debug('unknown response="%s"', line)
                    else:
                        self._logger.debug('line="%s"', line)
                        if callback:
                            if frame_handler:
                                callback(frame_handler(line))
                            else:
                                callback(line)
                        else:
                            self._logger.info('No callback set?')

        except socket.error as sock_err:
            self._logger.exception(sock_err)
            raise


class UDP(APRS):

    """APRS-IS UDP Class."""

    def __init__(self, user, password='-1', server=None, port=None):
        super(UDP, self).__init__(user, password)
        server = server or aprs.APRSIS_SERVERS[0]
        port = port or aprs.APRSIS_RX_PORT
        self._addr = (server, int(port))
        self.use_i_construct = True

    def start(self):
        """
        Connects & logs in to APRS-IS.
        """
        self.interface = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

    def send(self, frame):
        """
        Sends frame to APRS-IS.

        :param frame: Frame to send to APRS-IS.
        :type frame: str
        """
        self._logger.info('Sending frame="%s"', frame)
        content = b"\n".join([self._auth, str(frame)])
        return self.interface.sendto(content, self._addr)


class HTTP(APRS):

    """APRS-IS HTTP Class."""

    def __init__(self, user: bytes, password: bytes=b'-1', url: bytes=b'',
                 headers=None) -> None:
        super(HTTP, self).__init__(user, password)
        self.url = url or aprs.APRSIS_URL
        self.headers = headers or aprs.APRSIS_HTTP_HEADERS
        self.use_i_construct = True

    def start(self):
        """
        Connects & logs in to APRS-IS.
        """
        self.interface = requests.post

    def send(self, frame: bytes) -> bool:
        """
        Sends frame to APRS-IS.

        :param frame: Frame to send to APRS-IS.
        :type frame: str
        """
        if isinstance(frame, str):
            frame = aprs.parse_frame(frame)
        if isinstance(frame, aprs.Frame):
            frame = bytes(frame)
        self._logger.info('Sending frame="%s"', frame)
        content = b"\n".join([self._auth, frame])
        result = self.interface(self.url, data=content, headers=self.headers)
        return result.status_code == 204


class InformationField(object):

    """
    Class for APRS 'Information' Field.
    """

    _logger = logging.getLogger(__name__)  # pylint: disable=R0801
    if not _logger.handlers:  # pylint: disable=R0801
        _logger.setLevel(aprs.LOG_LEVEL)  # pylint: disable=R0801
        _console_handler = logging.StreamHandler()  # pylint: disable=R0801
        _console_handler.setLevel(aprs.LOG_LEVEL)  # pylint: disable=R0801
        _console_handler.setFormatter(aprs.LOG_FORMAT)  # pylint: disable=R0801
        _logger.addHandler(_console_handler)  # pylint: disable=R0801
        _logger.propagate = False  # pylint: disable=R0801

    __slots__ = ['data_type', 'data', 'safe']

    def __init__(self, data: bytes=b'', data_type: bytes=b'undefined',
                 safe: bool=False) -> None:
        self.data = data
        self.data_type = data_type
        self.safe = safe

    def __repr__(self) -> str:
        if self.safe:
            try:
                decoded_data = self.data.decode('UTF-8')
            except UnicodeDecodeError as ex:
                decoded_data = self.data.decode('UTF-8', 'backslashreplace')
            return decoded_data
        else:
            return self.data.decode()

    def __bytes__(self) -> bytes:
        return self.data


class PositionFrame(Frame):

    __slots__ = ['lat', 'lng', 'source', 'destination', 'path', 'table',
                 'symbol', 'comment', 'ambiguity']

    _logger = logging.getLogger(__name__)  # pylint: disable=R0801
    if not _logger.handlers:  # pylint: disable=R0801
        _logger.setLevel(aprs.LOG_LEVEL)  # pylint: disable=R0801
        _console_handler = logging.StreamHandler()  # pylint: disable=R0801
        _console_handler.setLevel(aprs.LOG_LEVEL)  # pylint: disable=R0801
        _console_handler.setFormatter(aprs.LOG_FORMAT)  # pylint: disable=R0801
        _logger.addHandler(_console_handler)  # pylint: disable=R0801
        _logger.propagate = False  # pylint: disable=R0801

    def __init__(self, source: bytes, destination: bytes, path: typing.List,
                 table: bytes, symbol: bytes, comment: bytes, lat: float,
                 lng: float, ambiguity: float) -> None:
        self.table = table
        self.symbol = symbol
        self.comment = comment
        self.lat = lat
        self.lng = lng
        self.ambiguity = ambiguity
        info = self.create_info_field()
        super(PositionFrame, self).__init__(source, destination, path, info)

    def create_info_field(self) -> bytes:
        enc_lat = aprs.dec2dm_lat(self.lat)
        enc_lat_amb = bytes(aprs.ambiguate(enc_lat, self.ambiguity), 'UTF-8')
        enc_lng = aprs.dec2dm_lng(self.lng)
        enc_lng_amb = bytes(aprs.ambiguate(enc_lng, self.ambiguity), 'UTF-8')
        frame = [
            b'=',
            enc_lat_amb,
            self.table,
            enc_lng_amb,
            self.symbol,
            self.comment
        ]
        return b''.join(frame)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Python APRS KISS Module Class Definitions."""

import logging

import kiss  # pylint: disable=R0801

import aprs  # pylint: disable=R0801

__author__ = 'Greg Albrecht W2GMD <oss@undef.net>'  # NOQA pylint: disable=R0801
__copyright__ = 'Copyright 2017 Greg Albrecht and Contributors'  # NOQA pylint: disable=R0801
__license__ = 'Apache License, Version 2.0'  # NOQA pylint: disable=R0801


class Frame(object):

    """
    Frame Class.

    Defines the components of an APRS Frame and can decode a frame
    from either ASCII or KISS.
    """

    __slots__ = ['frame', 'source', 'destination', 'path', 'text']

    _logger = logging.getLogger(__name__)  # pylint: disable=R0801
    if not _logger.handlers:  # pylint: disable=R0801
        _logger.setLevel(aprs.LOG_LEVEL)  # pylint: disable=R0801
        _console_handler = logging.StreamHandler()  # pylint: disable=R0801
        _console_handler.setLevel(aprs.LOG_LEVEL)  # pylint: disable=R0801
        _console_handler.setFormatter(aprs.LOG_FORMAT)  # pylint: disable=R0801
        _logger.addHandler(_console_handler)  # pylint: disable=R0801
        _logger.propagate = False  # pylint: disable=R0801

    def __init__(self, frame=None):
        self.source = ''
        self.destination = 'APRS'
        self.path = []
        self.text = ''
        if frame is not None:
            self.frame = kiss.strip_df_start(str(frame))
            self.parse()

    def __repr__(self):
        full_path = [str(self.destination)]
        full_path.extend([str(p) for p in self.path])
        frame = "%s>%s:%s" % (
            self.source,
            ','.join(full_path),
            self.text
        )
        return frame

    def to_h(self):
        """
        Returns an Frame as a Hex String.
        """
        return str(self).encode('hex')

    def parse(self, frame=None):
        """
        Parses an Frame from either ASCII or KISS Encoded frame.
        """
        # Allows to be called as class method:
        if frame is not None:
            self.frame = frame

        try:
            self.parse_kiss()
        except IndexError as exc:
            self._logger.info('Not a KISS Frame? %s', self.frame.encode('hex'))

        if not self.source or not self.destination:
            try:
                self.parse_text()
            except UnicodeDecodeError as exc:
                self._logger.info(
                    'Cannot decode frame=%s', self.frame.encode('hex'))
                self._logger.exception(exc)

    def parse_text(self):
        """
        Parses and Extracts the components of an ASCII-Encoded Frame.
        """
        frame_so_far = ''

        for char in self.frame:
            if '>' in char and not self.source:
                self.source = aprs.Callsign(frame_so_far)
                frame_so_far = ''
            elif ':' in char:
                if not self.path:
                    if ',' in frame_so_far:
                        self.destination = aprs.Callsign(
                            frame_so_far.split(',')[0])
                        self.path = []
                        for path in frame_so_far.split(',')[1:]:
                            self.path.append(aprs.Callsign(path))
                        frame_so_far = ''
                    elif not self.destination:
                        self.destination = aprs.Callsign(frame_so_far)
                        frame_so_far = ''
                    else:
                        frame_so_far = ''.join([frame_so_far, char])
                else:
                    frame_so_far = ''.join([frame_so_far, char])
            else:
                frame_so_far = ''.join([frame_so_far, char])

        self.text = frame_so_far

    def parse_kiss(self):
        """
        Parses and Extracts the components of an KISS-Encoded Frame.
        """
        frame_len = len(self.frame)

        if frame_len < 16:
            self._logger.debug('Frame len(%s) < 16, Exiting.', frame_len)
            return

        for raw_slice in range(0, frame_len):

            # Is address field length correct?
            # Find the first ODD Byte followed by the next boundary:
            if (ord(self.frame[raw_slice]) & 0x01
                    and ((raw_slice + 1) % 7) == 0):

                i = (raw_slice + 1) / 7

                # Less than 2 callsigns?
                if 1 < i < 11:
                    # For frames <= 70 bytes
                    if frame_len >= raw_slice + 2:
                        if (ord(self.frame[raw_slice + 1]) & 0x03 == 0x03 and
                                ord(self.frame[raw_slice + 2]) in
                                [0xf0, 0xcf]):
                            self._extract_kiss_text(raw_slice)
                            self._extract_kiss_destination()
                            self._extract_kiss_source()
                            self._extract_kiss_path(i)

    def encode_kiss(self):
        """
        Encodes an Frame as KISS.
        """
        enc_frame = ''.join([
            self.destination.encode_kiss(),
            self.source.encode_kiss(),
            ''.join([path_call.encode_kiss() for path_call in self.path])
        ])
        return ''.join([
            enc_frame[:-1],
            chr(ord(enc_frame[-1]) | 0x01),
            kiss.SLOT_TIME,
            chr(0xF0),
            self.text
        ])

    def _extract_kiss_text(self, raw_slice):
        """
        Extracts a Text portion of a KISS-Encoded Frame.
        """
        self.text = self.frame[raw_slice + 3:]

    def _extract_kiss_source(self):
        """
        Extracts a Source Callsign of a KISS-Encoded Frame.
        """
        self.source = aprs.Callsign(self.frame[7:])

    def _extract_kiss_destination(self):
        """
        Extracts a Destination Callsign of a KISS-Encoded Frame.
        """
        self.destination = aprs.Callsign(self.frame)

    def _extract_kiss_path(self, start):
        """
        Extracts path from raw APRS KISS frame.
        """
        for i in range(2, start):
            path_call = aprs.Callsign(self.frame[i * 7:])

            if path_call:
                if ord(self.frame[i * 7 + 6]) & 0x80:
                    path_call.digi = True

                self.path.append(path_call)


class SerialKISS(kiss.SerialKISS):

    """APRS interface for KISS serial devices."""

    def __init__(self, port, speed, strip_df_start=False):
        super(SerialKISS, self).__init__(port, speed, strip_df_start)
        self.send = self.write
        self.receive = self.read
        self.use_i_construct = False

    def write(self, frame):
        """Writes APRS-encoded frame to KISS device.

        :param frame: APRS frame to write to KISS device.
        :type frame: str
        """
        super(SerialKISS, self).write(frame.encode_kiss())


class TCPKISS(kiss.TCPKISS):

    """APRS interface for KISS serial devices."""

    def __init__(self, host, port, strip_df_start=False):
        super(TCPKISS, self).__init__(host, port, strip_df_start)
        self.send = self.write
        self.receive = self.read
        self.use_i_construct = False

    def write(self, frame):
        """
        Writes APRS-encoded frame to KISS device.

        :param frame: APRS frame to write to KISS device.
        :type frame: str
        """
        super(TCPKISS, self).write(frame.encode_kiss())



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
PyDecimalDegrees - geographic coordinates conversion utility.

Copyright (C) 2006-2013 by Mateusz Łoskot <mateusz@loskot.net>
Copyright (C) 2010-2013 by Evan Wheeler <ewheeler@unicef.org>

This file is part of PyDecimalDegrees module.

This software is provided 'as-is', without any express or implied warranty.
In no event will the authors be held liable for any damages arising from
the use of this software.

Permission is granted to anyone to use this software for any purpose,
including commercial applications, and to alter it and redistribute it freely,
subject to the following restrictions:
1. The origin of this software must not be misrepresented; you must not
   claim that you wrote the original software. If you use this software
   in a product, an acknowledgment in the product documentation would be
   appreciated but is not required.
2. Altered source versions must be plainly marked as such, and must not be
   misrepresented as being the original software.
3. This notice may not be removed or altered from any source distribution.

DESCRIPTION

DecimalDegrees module provides functions to convert between
degrees/minutes/seconds and decimal degrees.

Original source distribution:
http://mateusz.loskot.net/software/gis/pydecimaldegrees/

Inspired by Walter Mankowski's Geo::Coordinates::DecimalDegrees module
for Perl, originally located in CPAN Archives:
http://search.cpan.org/~waltman/Geo-Coordinates-DecimalDegrees-0.05/

doctest examples are based following coordinates:
DMS: 121 8' 6"
DM: 121 8.1'
DD: 121.135

To run doctest units just execut this module script as follows
(-v instructs Python to run script in verbose mode):

$ python decimaldegrees.py [-v]

"""

import decimal as libdecimal

from decimal import Decimal as D

__revision__ = '$Revision: 1.1 $'


def decimal2dms(decimal_degrees):
    """ Converts a floating point number of degrees to the equivalent
    number of degrees, minutes, and seconds, which are returned
    as a 3-element tuple of decimals. If 'decimal_degrees' is negative,
    only degrees (1st element of returned tuple) will be negative,
    minutes (2nd element) and seconds (3rd element) will always be positive.

    Example:

        >>> decimal2dms(121.135)
        (Decimal('121'), Decimal('8'), Decimal('6.000'))
        >>> decimal2dms(-121.135)
        (Decimal('-121'), Decimal('8'), Decimal('6.000'))

    """

    degrees = D(int(decimal_degrees))
    decimal_minutes = libdecimal.getcontext().multiply(
        (D(str(decimal_degrees)) - degrees).copy_abs(), D(60))
    minutes = D(int(decimal_minutes))
    seconds = libdecimal.getcontext().multiply(
        (decimal_minutes - minutes), D(60))
    return (degrees, minutes, seconds)


def decimal2dm(decimal_degrees):
    """
    Converts a floating point number of degrees to the degress & minutes.

    Returns a 2-element tuple of decimals.

    If 'decimal_degrees' is negative, only degrees (1st element of returned
    tuple) will be negative, minutes (2nd element) will always be positive.

    Example:

        >>> decimal2dm(121.135)
        (Decimal('121'), Decimal('8.100'))
        >>> decimal2dm(-121.135)
        (Decimal('-121'), Decimal('8.100'))

    """
    degrees = D(int(decimal_degrees))

    minutes = libdecimal.getcontext().multiply(
        (D(str(decimal_degrees)) - degrees).copy_abs(), D(60))

    return (degrees, minutes)


def dms2decimal(degrees, minutes, seconds):
    """ Converts degrees, minutes, and seconds to the equivalent
    number of decimal degrees. If parameter 'degrees' is negative,
    then returned decimal-degrees will also be negative.

    NOTE: this method returns a decimal.Decimal

    Example:

        >>> dms2decimal(121, 8, 6)
        Decimal('121.135')
        >>> dms2decimal(-121, 8, 6)
        Decimal('-121.135')

    """
    decimal = D(0)
    degs = D(str(degrees))
    mins = libdecimal.getcontext().divide(D(str(minutes)), D(60))
    secs = libdecimal.getcontext().divide(D(str(seconds)), D(3600))

    if degrees >= D(0):
        decimal = degs + mins + secs
    else:
        decimal = degs - mins - secs

    return libdecimal.getcontext().normalize(decimal)


def dm2decimal(degrees, minutes):
    """ Converts degrees and minutes to the equivalent number of decimal
    degrees. If parameter 'degrees' is negative, then returned decimal-degrees
    will also be negative.

    Example:

        >>> dm2decimal(121, 8.1)
        Decimal('121.135')
        >>> dm2decimal(-121, 8.1)
        Decimal('-121.135')

    """
    return dms2decimal(degrees, minutes, 0)


def run_doctest():  # pragma: no cover
    """Runs doctests for this module."""
    import doctest
    return doctest.testmod()


if __name__ == '__main__':
    run_doctest()  # pragma: no cover



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Python APRS Module Geo Utility Function Definitions."""

import aprs.decimaldegrees

__author__ = 'Greg Albrecht W2GMD <oss@undef.net>'  # NOQA pylint: disable=R0801
__copyright__ = 'Copyright 2017 Greg Albrecht and Contributors'  # NOQA pylint: disable=R0801
__license__ = 'Apache License, Version 2.0'  # NOQA pylint: disable=R0801


def dec2dm_lat(dec: float) -> str:
    """
    Converts DecDeg to APRS Coord format.

    See: http://ember2ash.com/lat.htm

    Source: http://stackoverflow.com/questions/2056750

    Example:
        >>> test_lat = 37.7418096
        >>> aprs_lat = dec2dm_lat(test_lat)
        >>> aprs_lat
        '3744.51N'
        >>> test_lat = -8.01
        >>> aprs_lat = dec2dm_lat(test_lat)
        >>> aprs_lat
        '0800.60S'
    """
    dec_min = aprs.decimaldegrees.decimal2dm(dec)

    deg = dec_min[0]
    abs_deg = abs(deg)

    if not deg == abs_deg:
        suffix = 'S'
    else:
        suffix = 'N'

    return "%02d%05.2f%s" % (abs_deg, dec_min[1], suffix)


def dec2dm_lng(dec: float) -> str:
    """
    Converts DecDeg to APRS Coord format.

    See: http://ember2ash.com/lat.htm

    Example:
        >>> test_lng = 122.38833
        >>> aprs_lng = dec2dm_lng(test_lng)
        >>> aprs_lng
        '12223.30E'
        >>> test_lng = -99.01
        >>> aprs_lng = dec2dm_lng(test_lng)
        >>> aprs_lng
        '09900.60W'
    """
    dec_min = aprs.decimaldegrees.decimal2dm(dec)

    deg = dec_min[0]
    abs_deg = abs(deg)

    if not deg == abs_deg:
        suffix = 'W'
    else:
        suffix = 'E'

    return "%03d%05.2f%s" % (abs_deg, dec_min[1], suffix)


def ambiguate(pos: float, ambiguity: int) -> str:
    """
    Adjust ambiguity of position.

    Derived from @asdil12's `process_ambiguity()`.

    >>> pos = '12345.67N'
    >>> ambiguate(pos, 0)
    '12345.67N'
    >>> ambiguate(pos, 1)
    '12345.6 N'
    >>> ambiguate(pos, 2)
    '12345.  N'
    >>> ambiguate(pos, 3)
    '1234 .  N'
    """
    num = bytearray(pos, 'UTF-8')
    for i in range(0, ambiguity):
        if i > 1:
            # skip the dot
            i += 1
        # skip the direction
        i += 2
        num[-i] = ord(' ')
    return num.decode()


def run_doctest():  # pragma: no cover
    """Runs doctests for this module."""
    import doctest
    return doctest.testmod()


if __name__ == '__main__':
    run_doctest()  # pragma: no cover



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Python APRS Module Function Definitions."""

import typing

import aprs  # pylint: disable=R0801

__author__ = 'Greg Albrecht W2GMD <oss@undef.net>'  # NOQA pylint: disable=R0801
__copyright__ = 'Copyright 2017 Greg Albrecht and Contributors'  # NOQA pylint: disable=R0801
__license__ = 'Apache License, Version 2.0'  # NOQA pylint: disable=R0801


AprsCallsign = typing.TypeVar('AprsCallsign', bound='aprs.Callsign')
AprsFrame = typing.TypeVar('AprsFrame', bound='aprs.Frame')


def parse_frame(raw_frame: typing.Union[bytes, str]) -> AprsFrame:
    """
    Parses an AX.25/APRS Frame from either plain-text or AX.25.
    """
    if isinstance(raw_frame, aprs.Frame):
        return raw_frame
    elif isinstance(raw_frame, str):
        return parse_frame_text(bytes(raw_frame, 'UTF-8'))
    elif isinstance(raw_frame, bytes) or isinstance(raw_frame, bytearray):
        if aprs.ADDR_INFO_DELIM in raw_frame:
            return parse_frame_ax25(raw_frame)
        else:
            return parse_frame_text(raw_frame)


def parse_frame_text(raw_frame: bytes) -> AprsFrame:
    """
    Parses and Extracts the components of a str Frame.
    """
    parsed_frame = aprs.Frame()
    _path = []

    # Source>Destination
    sd_delim = raw_frame.index(b'>')

    parsed_frame.set_source(raw_frame[:sd_delim])

    # Path:Info
    pi_delim = raw_frame.index(b':')

    parsed_path = raw_frame[sd_delim + 1:pi_delim]
    if b',' in parsed_path:
        for path in parsed_path.split(b','):
            _path.append(path)
        parsed_frame.set_destination(_path.pop(0))
        parsed_frame.set_path(_path)
    else:
        parsed_frame.set_destination(parsed_path)

    parsed_frame.set_info(raw_frame[pi_delim + 1:])

    return parsed_frame


def parse_frame_ax25(raw_frame: bytes) -> AprsFrame:
    """
    Parses and Extracts the components of an AX.25-Encoded Frame.
    """
    parsed_frame = aprs.Frame()
    kiss_call = False

    _frame = raw_frame.strip(aprs.AX25_FLAG)
    if (_frame.startswith(aprs.KISS_DATA_FRAME) or
            _frame.endswith(aprs.KISS_DATA_FRAME)):
        _frame = _frame.lstrip(aprs.KISS_DATA_FRAME)
        _frame = _frame.rstrip(aprs.KISS_DATA_FRAME)
        kiss_call = True

    # Use these two fields as the address/information delimiter
    frame_addressing, frame_information = _frame.split(aprs.ADDR_INFO_DELIM)

    info_field = frame_information.rstrip(b'\xFF\x07')

    destination = parse_callsign_ax25(frame_addressing, kiss_call)
    source = parse_callsign_ax25(frame_addressing[7:], kiss_call)

    paths = frame_addressing[7+7:]
    n_paths = int(len(paths) / 7)
    n = 0
    path = []
    while n < n_paths:
        path.append(parse_callsign_ax25(paths[:7]))
        paths = paths[7:]
        n += 1

    parsed_frame.set_source(source)
    parsed_frame.set_destination(destination)
    parsed_frame.set_path(path)
    parsed_frame.set_info(info_field)
    return parsed_frame


def parse_callsign(raw_callsign: bytes) -> AprsCallsign:
    """
    Parses an AX.25/APRS Callsign from plain-text or AX.25 input.
    """
    if isinstance(raw_callsign, aprs.Callsign):
        return raw_callsign
    try:
        return parse_callsign_ax25(raw_callsign)
    except:
        if isinstance(raw_callsign, str):
            return parse_callsign_text(bytes(raw_callsign, 'UTF-8'))
        else:
            return parse_callsign_text(raw_callsign)


def parse_callsign_text(raw_callsign: bytes) -> AprsCallsign:
    """
    Parses an AX.25/APRS Callsign & SSID from a plain-text AX.25/APRS Frame.
    """
    parsed_callsign: AprsCallsign = aprs.Callsign()
    _callsign = raw_callsign
    ssid = b'0'
    digi = False

    if b'*' in _callsign:
        _callsign = _callsign.strip(b'*')
        digi = True

    if b'-' in _callsign:
        _callsign, ssid = _callsign.split(b'-')

    parsed_callsign.set_callsign(_callsign)
    parsed_callsign.set_ssid(ssid)
    parsed_callsign.set_digi(digi)

    return parsed_callsign


def parse_callsign_ax25(raw_callsign: bytes, kiss_call: bool=False) -> AprsCallsign:
    """
    Extracts a Callsign and SSID from a AX.25 Encoded APRS Frame.

    :param frame: AX.25 Encoded APRS Frame.
    :type frame: str
    """
    parsed_callsign = aprs.Callsign()
    _callsign = bytes()
    digi = False

    for _chunk in raw_callsign[:6]:
        chunk = _chunk & 0xFF
        if chunk & 1:
            # aprx: /* Bad address-end flag ? */
            raise aprs.BadCallsignError('Bad address-end flag.')

        # Shift by one bit:
        chunk = chunk >> 1
        chr_chunk = chr(chunk)

        if chr_chunk.isalnum():
            _callsign += bytes([chunk])

    # 7th byte carries SSID or digi:
    seven_chunk = raw_callsign[6] & 0xFF
    ssid = (seven_chunk >> 1) & 0x0F  # Limit it to 4 bits.

    # FIXME gba@20170809: This works for KISS frames, but not otherwise.
    # Should consult: https://github.com/chrissnell/GoBalloon/blob/master/ax25/encoder.go
    if kiss_call:
        if seven_chunk >> 1 & 0x80:
            digi = True
    else:
        if seven_chunk & 0x80:
            digi = True

    parsed_callsign.set_callsign(_callsign)
    parsed_callsign.set_ssid(ssid)
    parsed_callsign.set_digi(digi)

    return parsed_callsign


def parse_info_field(raw_data: bytes, handler=None) -> bytes:
    if not raw_data:
        return bytes()
    elif isinstance(raw_data, aprs.InformationField):
        return raw_data
    elif isinstance(raw_data, bytes) or isinstance(raw_data, bytearray):
        data_type = b''
        data_type_field = chr(raw_data[0])
        data_type = aprs.DATA_TYPE_MAP.get(data_type_field)

        if data_type:
            if handler:
                handler_func = getattr(
                    handler,
                    "handle_data_type_%s" % data_type,
                    None
                )
                return handler_func(raw_data, data_type)

        return aprs.InformationField(raw_data, data_type, safe=True)


def default_data_handler(data: bytes, data_type: bytes) -> bytes:
    """
    Handler for Undefined Data Types.
    """
    try:
        decoded_data = data.decode('UTF-8')
    except UnicodeDecodeError as ex:
        decoded_data = data.decode('UTF-8', 'backslashreplace')

    return aprs.InformationField(decoded_data, data_type)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Python APRS Module Utility Functions Definitions."""

__author__ = 'Greg Albrecht W2GMD <oss@undef.net>'  # NOQA pylint: disable=R0801
__copyright__ = 'Copyright 2017 Greg Albrecht and Contributors'  # NOQA pylint: disable=R0801
__license__ = 'Apache License, Version 2.0'  # NOQA pylint: disable=R0801


def valid_callsign(callsign):
    """
    Validates an over-the-air callsign. APRS-IS is more forgiving.

    Verifies that a valid callsign is valid:
    >>> valid_callsign('W2GMD-1')
    True
    >>>

    Verifies that an invalid callsign is invalid:
    >>> valid_callsign('BURRITOS-99')
    False
    >>>

    :param callsign: Callsign to validate.
    :type callsign: str

    :returns: True if valid, False otherwise.
    :rtype: bool
    """
    callsign = callsign.lstrip().rstrip().strip('*')

    if '-' in callsign:
        if not callsign.count('-') == 1:
            return False
        else:
            callsign, ssid = callsign.split('-')
    else:
        ssid = 0

    # Test length, call should be <6.
    if (len(callsign) < 2 or len(callsign) > 6 or len(str(ssid)) < 1 or
            len(str(ssid)) > 2):
        return False

    for char in callsign:
        if not (char.isalpha() or char.isdigit()):
            if char == '*' and callsign[-1] == '*':
                next
            else:
                return False

    if not str(ssid).isdigit():
        return False

    if int(ssid) < 0 or int(ssid) > 15:
        return False

    return True


def run_doctest():  # pragma: no cover
    """Runs doctests for this module."""
    import doctest
    return doctest.testmod()


if __name__ == '__main__':
    run_doctest()  # pragma: no cover
